from django.test import SimpleTestCase

from django_kafka.utils.message import Header


class HeaderTestCase(SimpleTestCase):
    def test_get(self):
        headers = [("header", "abc"), ("header", "def")]

        self.assertEqual(Header.get(headers, "header"), "abc")
        self.assertEqual(Header.get(None, "header"), None)

    def test_list(self):
        headers = [("header", "abc"), ("header", "def")]

        self.assertEqual(Header.list(headers, "header"), ["abc", "def"])
        self.assertEqual(Header.list(None, "header"), [])



from datetime import datetime, timedelta
from typing import TYPE_CHECKING

from django.core.exceptions import ObjectDoesNotExist
from django.db import IntegrityError
from django.utils import timezone

if TYPE_CHECKING:
    from django_kafka.topic import TopicConsumer


class RetrySettings:
    relational_errors = (
        ObjectDoesNotExist,
        IntegrityError,
    )

    def __init__(
        self,
        max_retries: int,
        delay: int,
        backoff: bool = False,
        include: list[type[Exception]] | None = None,
        exclude: list[type[Exception]] | None = None,
        blocking: bool = True,
        log_every: int | None = None,
        use_offset_tracker: bool = False,
    ):
        """
        :param max_retries: maximum number of retry attempts (use -1 for infinite)
        :param delay: delay (seconds)
        :param backoff: use an exponential backoff delay
        :param include: exception types to retry for
        :param exclude: exception types to exclude from retry
        :param blocking: block the consumer process during retry
        :param log_every: log every Nth retry attempt, default is not to log
        :param use_offset_tracker: use the offset tracker to skip failing messages
        """
        if max_retries < -1:
            raise ValueError("max_retries must be greater than -1")
        if delay <= 0:
            raise ValueError("delay must be greater than zero")
        if include is not None and exclude is not None:
            raise ValueError("cannot specify both include and exclude")

        self.max_retries = max_retries
        self.delay = delay
        self.backoff = backoff
        self.include = include
        self.exclude = exclude
        self.blocking = blocking
        self.log_every = log_every
        self.use_offset_tracker = use_offset_tracker

    def __call__(self, topic_cls: type["TopicConsumer"]):
        topic_cls.retry_settings = self
        return topic_cls

    def attempts_exceeded(self, attempt):
        if self.max_retries == -1:
            return False
        return attempt > self.max_retries

    def skip_by_offset(self, msg, exc: Exception):
        # avoiding import errors
        from django_kafka.models import KeyOffsetTracker  # Apps aren't loaded yet

        return all(
            [
                self.use_offset_tracker,
                isinstance(exc, self.relational_errors),
                KeyOffsetTracker.objects.has_future_offset(msg),
            ],
        )

    def should_retry(self, msg, attempt: int, exc: Exception) -> bool:
        if self.skip_by_offset(msg, exc):
            return False
        if self.attempts_exceeded(attempt):
            return False
        if self.include is not None:
            return any(isinstance(exc, e) for e in self.include)
        if self.exclude is not None:
            return not any(isinstance(exc, e) for e in self.exclude)
        return True

    def should_log(self, attempt: int):
        """returns if a message attempt failure should be logged"""
        if self.log_every is None:
            return False
        return attempt % self.log_every == 0

    def get_retry_delay(self, attempt: int) -> int:
        return self.delay * (2 ** (attempt - 1)) if self.backoff else self.delay

    def get_retry_time(self, attempt: int) -> datetime:
        return timezone.now() + timedelta(seconds=self.get_retry_delay(attempt))



from django.contrib import admin

from django_kafka.models import KeyOffsetTracker


@admin.register(KeyOffsetTracker)
class KeyOffsetTrackerAdmin(admin.ModelAdmin):
    list_display = ("topic", "key", "offset")
    search_fields = ("topic", "key")



from django_kafka.exceptions import DjangoKafkaError
from django_kafka.topic import TopicConsumer


class Topics:
    _topic_consumers: tuple["TopicConsumer", ...]
    _match: dict[str, "TopicConsumer"]

    def __init__(self, *topic_consumers: "TopicConsumer"):
        self._topic_consumers = topic_consumers
        self._match: dict[str, TopicConsumer] = {}

    def get(self, topic_name: str) -> "TopicConsumer":
        if topic_name not in self._match:
            topic_consumer = next((t for t in self if t.matches(topic_name)), None)
            if not topic_consumer:
                raise DjangoKafkaError(f"No topic registered for `{topic_name}`")
            self._match[topic_name] = topic_consumer

        return self._match[topic_name]

    def get_retryable(self, blocking=False) -> list["TopicConsumer"]:
        return [
            topic
            for topic in self._topic_consumers
            if topic.retry_settings and topic.retry_settings.blocking == blocking
        ]

    @property
    def names(self) -> list[str]:
        return [topic.name for topic in self]

    def __iter__(self):
        yield from self._topic_consumers

    def __bool__(self) -> bool:
        return bool(self.names)



from datetime import UTC, datetime
from enum import IntEnum

MessageHeaders = list[tuple[str, str]]


class Header:
    @classmethod
    def list(cls, headers: MessageHeaders | None, header: str) -> list[str]:
        """returns all occurrences for the header"""
        if headers is not None:
            return [v for k, v in headers if k == header]
        return []

    @classmethod
    def get(cls, headers: MessageHeaders | None, header: str) -> str | None:
        """returns the first encountered value for the header"""
        if headers is not None:
            return next((v for k, v in headers if k == header), None)
        return None


class MessageTimestamp(IntEnum):
    """
    https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#confluent_kafka.Message.timestamp
    Values are taken from the constants from `confluent_kafka.cimpl.py`
    """

    NOT_AVAILABLE = 0
    CREATE_TIME = 1
    LOG_APPEND_TIME = 2

    @classmethod
    def to_datetime(cls, msg_timestamp: [int, int]) -> datetime | None:
        timestamp_type, timestamp = msg_timestamp

        if timestamp_type == MessageTimestamp.NOT_AVAILABLE:
            return None

        return datetime.fromtimestamp(timestamp / 1000, UTC)



import logging
import traceback
from datetime import datetime
from pydoc import locate
from typing import TYPE_CHECKING, Optional

from confluent_kafka import Consumer as ConfluentConsumer

from django_kafka.conf import settings

from .managers import PauseManager, RetryManager
from .topics import Topics

if TYPE_CHECKING:
    from confluent_kafka import cimpl

    from django_kafka.retry.settings import RetrySettings
    from django_kafka.topic import TopicConsumer

logger = logging.getLogger(__name__)


class Consumer:
    """
    Available settings of the producers (P) and consumers (C)
        https://github.com/confluentinc/librdkafka/blob/master/CONFIGURATION.md
    Consumer configs
        https://kafka.apache.org/documentation/#consumerconfigs
    Kafka Client Configuration
        https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#kafka-client-configuration
    confluent_kafka.Consumer API
        https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#pythonclient-consumer
    """

    topics: Topics
    config: dict

    polling_freq = settings.POLLING_FREQUENCY
    default_logger = logger
    default_error_handler = settings.ERROR_HANDLER

    def __init__(self):
        self.config = self.build_config()
        self._consumer = ConfluentConsumer(self.config)
        self._retries = RetryManager()  # blocking retry manager
        self._pauses = PauseManager()

    def __getattr__(self, name):
        """proxy consumer methods."""
        return getattr(self._consumer, name)

    @classmethod
    def build_config(cls):
        return {
            "client.id": settings.CLIENT_ID,
            **settings.GLOBAL_CONFIG,
            **settings.CONSUMER_CONFIG,
            "logger": cls.default_logger,
            "error_cb": locate(cls.default_error_handler)(),
            **getattr(cls, "config", {}),
        }

    @property
    def group_id(self) -> str:
        return self.config["group.id"]

    def commit_offset(self, msg: "cimpl.Message"):
        if not self.config.get("enable.auto.offset.store"):
            # Store the offset associated with msg to a local cache.
            # Stored offsets are committed to Kafka by a background
            #  thread every 'auto.commit.interval.ms'.
            # Explicitly storing offsets after processing gives at-least once semantics.
            self.store_offsets(msg)

    def pause_partition(self, msg: "cimpl.Message", until: datetime):
        """pauses message partition to process the message at a later time

        note: pausing is only retained within the python consumer class, and is not
        retained between separate consumer processes.
        """
        partition = self._pauses.set(msg, until)
        self.seek(partition)  # seek back to message offset to re-poll on unpause
        self.pause([partition])

    def resume_partitions(self):
        """resumes any paused partitions that are now ready"""
        for partition in self._pauses.pop_ready():
            self.resume([partition])

    def blocking_retry(
        self,
        retry_settings: "RetrySettings",
        msg: "cimpl.Message",
        exc: Exception,
    ) -> bool:
        """
        blocking retry, managed within the same consumer process

        :return: whether the message will be retried
        """
        attempt = self._retries.next(msg)
        if retry_settings.should_retry(msg, attempt, exc):
            until = retry_settings.get_retry_time(attempt)
            self.pause_partition(msg, until)

            if retry_settings.should_log(attempt):
                self.log_error(msg, exc_info=exc)

            return True
        return False

    def non_blocking_retry(
        self,
        retry_settings: "RetrySettings",
        msg: "cimpl.Message",
        exc: Exception,
    ):
        """
        non-blocking retry, managed by a separate topic and consumer process

        :return: whether the message will be retried
        """
        from django_kafka.retry.topic import RetryTopicProducer

        topic = RetryTopicProducer(
            retry_settings=retry_settings,
            group_id=self.group_id,
            msg=msg,
        )

        retried = topic.retry(exc=exc)

        if retried and retry_settings.should_log(topic.retry_attempt):
            self.log_error(msg, exc_info=exc)

        return retried

    def retry_msg(self, msg: "cimpl.Message", exc: Exception) -> (bool, bool):
        """
        :return tuple: The first element indicates if the message was retried and the
        second indicates if the retry was blocking.
        """
        retry_settings = self.get_topic(msg).retry_settings
        if not retry_settings:
            return False, False
        if retry_settings.blocking:
            return self.blocking_retry(retry_settings, msg, exc), True
        return self.non_blocking_retry(retry_settings, msg, exc), False

    def dead_letter_msg(self, msg: "cimpl.Message", exc: Exception):
        """publishes a message to the dead letter topic, with exception details"""
        from django_kafka.dead_letter.topic import DeadLetterTopicProducer

        DeadLetterTopicProducer(group_id=self.group_id, msg=msg).produce_for(
            header_summary=str(exc),
            header_detail=traceback.format_exc(),
        )

    def handle_exception(self, msg: "cimpl.Message", exc: Exception) -> bool:
        """
        return bool: Indicates if the message was processed and offset can be committed.
        """
        retried, blocking = self.retry_msg(msg, exc)
        if not retried:
            self.dead_letter_msg(msg, exc)
            self.log_error(msg, exc_info=exc)
            return True
        return not blocking

    def get_topic(self, msg: "cimpl.Message") -> "TopicConsumer":
        return self.topics.get(topic_name=msg.topic())

    def log_error(
        self,
        msg: Optional["cimpl.Message"] = None,
        exc_info: bool | Exception = False,
    ):
        error = f"'{self.__class__.__module__}.{self.__class__.__name__} failed'"
        if msg:
            topic = self.get_topic(msg)
            error = (
                f"{error} on '{topic.__class__.__module__}.{topic.__class__.__name__}'"
            )

            if msg_error := msg.error():
                error = f"{error}\nMessage error: '{msg_error}'"

        logger.error(error, exc_info=exc_info)

    def consume(self, msg):
        self.get_topic(msg).consume(msg)

    def process_message(self, msg: "cimpl.Message"):
        if msg.error():
            self.log_error(msg)
            return

        try:
            self.consume(msg)
        except Exception as exc:
            # ruff: noqa: BLE001 (do not stop consumer if message consumption fails in any circumstances)
            processed = self.handle_exception(msg, exc)
        else:
            processed = True

        if processed:
            self.commit_offset(msg)

    def poll(self) -> Optional["cimpl.Message"]:
        # poll for self.polling_freq seconds
        return self._consumer.poll(timeout=self.polling_freq)

    def start(self):
        self.subscribe(topics=self.topics.names)

    def run(self):
        try:
            self.start()
            while True:
                self.resume_partitions()
                if (msg := self.poll()) is not None:
                    self.process_message(msg)
        except Exception as exc:
            self.log_error(exc_info=exc)
            raise
        finally:
            self.stop()

    def stop(self):
        self.close()



from typing import ClassVar

from django.db import models
from django.utils.translation import gettext_lazy as _

from django_kafka.utils.message import MessageTimestamp


class KeyOffsetTrackerQuerySet(models.QuerySet):
    def log_msg_offset(self, msg) -> tuple["KeyOffsetTracker", bool]:
        return self.update_or_create(
            topic=msg.topic(),
            key=msg.key(),
            defaults={
                "offset": msg.offset(),
                "timestamp": msg.timestamp(),
            },
        )

    def has_future_offset(self, msg) -> bool:
        return self.filter(
            topic=msg.topic(),
            key=msg.key(),
            offset__gt=msg.offset(),
        ).exists()


class KeyOffsetTracker(models.Model):
    topic = models.CharField(_("topic"), max_length=256)
    key = models.BinaryField(_("key"))
    offset = models.PositiveIntegerField(_("offset"))
    create_time = models.DateTimeField(_("create time"), null=True)
    log_append_time = models.DateTimeField(_("log append time"), null=True)

    objects = KeyOffsetTrackerQuerySet.as_manager()

    class Meta:
        verbose_name = _("key offset tracker")
        verbose_name_plural = _("key offsets tracker")
        indexes: ClassVar = [
            models.Index(
                fields=["topic", "key"],
                include=["offset"],
                name="%(app_label)s_key_offset_index",
            ),
        ]

    def __str__(self):
        return f"{self.topic}:{self.key}:{self.offset}"

    @property
    def timestamp(self):
        if self.log_append_time:
            return self.log_append_time
        return self.create_time or self.log_append_time

    @timestamp.setter
    def timestamp(self, value: [MessageTimestamp, int]):
        if not value:
            return

        timestamp_type = value[0]

        if timestamp_type == MessageTimestamp.CREATE_TIME:
            self.create_time = MessageTimestamp.to_datetime(value)
            self.log_append_time = None

        elif timestamp_type == MessageTimestamp.LOG_APPEND_TIME:
            self.create_time = None
            self.log_append_time = MessageTimestamp.to_datetime(value)

        else:
            self.create_time = None
            self.log_append_time = None
